/*  osdchat - on-screen display messaging service
 *  Copyright (c) 2003 by Mike Smelov  <mike-at-argoncorp-dot-com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  See LICENSE for details.
 *
 */

 /* osdchatd.c */

#include "osdchatd.h"
#include "common.h"
#include "version.h"

/* defined in config.c */
extern struct xosd_s xset;
extern struct pref_s pref;

/* defined in common.c */
extern char *prog_name;

int main(int argc, char **argv)
{
	int opt;

	set_prog_name(argv[0]);
	
	/* get user info */
	if ( (user_info = getpwuid(getuid())) == NULL)
		printDBG(LVL_FATERR, "getpwuid(): %m\n");

	snprintf(pid_fname, sizeof(pid_fname), PID_FILE, user_info->pw_uid);
		
	/* parse these first -- they do not alter internal behavior */
	opterr = 0;
	while ((opt = getopt(argc, argv, "Dhk")) != -1) {
		switch (opt) {
		case 'D':
			debug = 1;
			break;
		case 'h':
			print_usage();
		case 'k':
			srv_kill();
		case '?':
			break;
		}
	}
	
	if (debug == 1) {
		printDBG(LVL_DEBUG, "Debugging ON\n");
		printDBG(LVL_DEBUG, "Program Name: %s\n", prog_name);
		printDBG(LVL_DEBUG, "Home Dir: %s\n", user_info->pw_dir);
		printDBG(LVL_DEBUG, "User Name: %s\n", user_info->pw_name);
		printDBG(LVL_DEBUG, "PID File: %s\n", pid_fname);
	}

	/* parse config file + cmd line args */
	parse_config();
	parse_args(argc, argv);

	/* go into the background if debug not set*/
	if (debug == 0 && daemon(0, 0) == -1) 
		printDBG(LVL_FATERR, "daemon(): %m\n");

	/* handle signals */
	signal(SIGINT, srv_set_quit);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, srv_sighup);
	signal(SIGTERM, srv_set_quit);

	/* start server */
	srv_start(srv_setup());

	return EXIT_FAILURE;
}

/* Initilize server socket.
   Return socket file descriptor.
*/
int srv_setup()
{
	int srv_fd, on = 1;
	struct sockaddr_in srv_addr;
	char buf[BUFSIZE];
	FILE *f;
	time_t t;

	if ( (srv_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
		printDBG(LVL_FATAL, "socket(): %m");

	/* setup server address */
	memset(&srv_addr, 0, sizeof(srv_addr));
	srv_addr.sin_family = AF_INET;
	srv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	srv_addr.sin_port = htons(pref.serv_port);

	/* re-use socket address to eliminiate a TIME_WAIT */
	if ( setsockopt(srv_fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1)
		printDBG(LVL_FATAL, "setsockopt(SO_REUSEADDR): %m");

	if ( bind(srv_fd, (struct sockaddr *) &srv_addr, sizeof(srv_addr)) == -1 )
		printDBG(LVL_FATAL, "Bind to port %d failed: %m", pref.serv_port);

	/* set umask - prevent pid file tampering if previous
	   umask is not set properly
	*/
	umask(022);

	if ( listen(srv_fd, SOMAXCONN) == -1 )
		printDBG(LVL_FATAL, "listen(): %s", strerror(errno));

	/* set listening socket to non-blocking in case an RST is sent after select
	   indicates a read but before accept is called.
	*/
	if ( fcntl(srv_fd, F_SETFL, fcntl(srv_fd, F_GETFL, 0) | O_NONBLOCK) < 0)
		printDBG(LVL_FATAL, "Cannot set listening socket as non-blocking: %m");

	/* write current pid to pid file */
	f = safe_fopen(pid_fname, "w");
	if (f) {
		fprintf(f, "%d\n", getpid());
		fclose(f);
	} else {
		printDBG(LVL_SYSLOG, "safe_fopen(%s): %m", pid_fname);
	}

	printDBG(LVL_DEBUG, "Server fd: %d\n", srv_fd);
        printDBG(LVL_SYSLOG, "osdchatd "OC_VERSION" daemon started.");
        t = time(NULL);
        strftime(buf, BUFSIZE, "---{ osdchatd started %b %d %Y %X }---", localtime(&t));
        log_msgs(buf);

	return srv_fd;
}

/* Handle client connections */
void srv_start(int srv_fd)
{
	struct sockaddr_in client_addr;
	int i, client_fd, client_len;

	fd_set test_fds;
	int fd, nr;

	FD_ZERO(&read_fds);
	FD_SET(srv_fd, &read_fds);

	max_fd = srv_fd;
	num_clients = 0;
	cs_init();

	for (;;) {
		if (reload_config)
			srv_reload_config();

		if (quit)
			srv_quit();

		test_fds = read_fds;

		if ( (nr = select(max_fd + 1, &test_fds, NULL, NULL, NULL)) == -1 ) {
			if (errno == EINTR)		/* interrupted by a signal */
        			continue;

			printDBG(LVL_FATAL, "select(): %m");
		}

		/* scan all file descripters for reads */
		for (fd = 0; fd <= max_fd; fd++) {
			if ( !FD_ISSET(fd, &test_fds) ) {
				continue;
			}

			printDBG(LVL_DEBUG, "---\ncurrent fd: %d\tmaxfd: %d\n---\n", fd, max_fd);

			if (fd == srv_fd) {
				/* remove dangling clients that have not sent their prompt yet */
				if (num_clients >= pref.max_clients) {
					for (i = 0; i < max_fd; i++)
						if (cs_get_tm(i) == 0) {
							srv_disconnect_client(i);
							break;
						}
				}

				/* we have a new client */
				client_len = sizeof(client_addr);
				if ( (client_fd = accept(fd, (struct sockaddr *) &client_addr, &client_len)) < 0 ) {
					if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK)
						continue;

					printDBG(LVL_FATAL, "accept(): %m");
				}

				/* set non-blocking io */
				if ( fcntl(client_fd, F_SETFL, O_NONBLOCK) < 0) {
					printDBG(LVL_SYSLOG, "Cannot set client fd as non-blocking: %m");
					close(client_fd);
				}

				srv_add_client(client_fd, &client_addr);
			}
			else if (cs_get_prompt(fd) != NULL)
				srv_old_client(fd);
			else
				srv_add_prompt(fd);

			/* all reads processed, quit loop */
			if (--nr < 0)
				break;
		}
	}
}

/* non-block io send */
void srv_send(int s, const void *msg, size_t len)
{
	int sent;

	sent = send(s, msg, len, MSG_DONTWAIT);

	if (sent == -1) {
		printDBG(LVL_DEBUG, "send(): %m\n");
		srv_disconnect_client(s);
		return;
	}

	if (sent != len)
		printDBG(LVL_SYSLOG, "Failed to send complete message to peer, bytes xfer: %d\n", sent);
}

/* Handle new client */
void srv_add_client(int client_fd, struct sockaddr_in *client_addr)
{
	char prompt[BUFSIZE], ip[32] = {0};
	int format_len, xoff;
	msg_t client_msg = MSG_CONNOK;
	int blocked = 0;

	if ( inet_ntop(AF_INET, &client_addr->sin_addr, ip, sizeof(ip)) == NULL)
		printDBG(LVL_FATAL, "inet_ntop(): %m\n");

	printDBG(LVL_DEBUG, "Client connecting from %s.\n", ip);

	/* Number of clients has reached maximum value, close new connection */
	if (num_clients >= pref.max_clients) {
		printDBG(LVL_SYSLOG, "Maximum clients reached, disconnecting client from %s.", ip);
		client_msg = MSG_SERVFULL;
	}

	/* Check if client's IP has been blocked */
	if (pref.blocklist && cs_is_blocked(client_addr->sin_addr.s_addr)) {
		printDBG(LVL_SYSLOG, "IP %s blocked, disconnecting client.", ip);
		client_msg = MSG_CLIBLOCK;
		blocked = 1;
	}

	/* If osd is not initilized yet and it fails to initilize then the
	   specified or default display is not ready or is restricting access.
	   Write a response to client, then close the client socket.
	*/
	if ( (xoff = (osd == NULL && !init_xosd())) ) {
		printDBG(LVL_SYSLOG, "Display not initialized, disconnecting client from %s", ip);
		client_msg = MSG_NODISP;
	}

	srv_send(client_fd, &client_msg, sizeof(client_msg));

	if (pref.loud_block && blocked && !xoff) {
		snprintf(prompt, sizeof(prompt), "*** Connection attempted by blocked client from %s", ip);
		show_osd(prompt, " ***");
	}

	if (client_msg != MSG_CONNOK) {
		close(client_fd);
		return;
	}

	FD_SET(client_fd, &read_fds);
	if (client_fd > max_fd)
		max_fd = client_fd;

	num_clients++;
	cs_add_client(client_fd, client_addr);

	printDBG(LVL_DEBUG, "num_clients: %d  fd: %d  ip: %s\n", num_clients, client_fd, ip);

	/* write prompt format to client */
	format_len = strlen(pref.p_format)+1;
	printDBG(LVL_DEBUG, "format sent to client: %s (length: %d)\n", pref.p_format, format_len);
	srv_send(client_fd, (char *) pref.p_format, format_len);
}

/* Read in client's prompt */
void srv_add_prompt(int client_fd)
{
	char prompt[BUFSIZE] = {0};
	int nread;

	(void) ioctl(client_fd, FIONREAD, &nread);

	if (nread == 0) {
		srv_disconnect_client(client_fd);
		return;
	}

	/* read formatted prompt from client */
	if (read(client_fd, &prompt, BUFSIZE-1) == -1 && errno == EAGAIN)
		return;

	if (prompt == NULL)
		return;

	prompt[BUFSIZE-1] = NIL;
	cs_set_prompt(client_fd, prompt);
	cs_set_tm(client_fd, time(NULL));

	printDBG(LVL_DEBUG, "Client sent prompt: %s (length: %d)\n", prompt, strlen(prompt));

	if (pref.loud_connect)
		show_osd(cs_get_prompt(client_fd), "*** CONNECTED ***");
}

/* Disconnect client.  Remove from client list, adjust read_fds
set decrement client count, and adjust max_fd if necessary */
void srv_disconnect_client(int fd)
{
	FD_CLR(fd, &read_fds);

	if (fd >= max_fd)
		max_fd--;

	printDBG(LVL_DEBUG, "close fd: %d\n", fd);
	cs_del_client(fd);
	num_clients--;
	(void) close(fd);
}

/* Handle known client */
void srv_old_client(int fd)
{
	char msg[BUFSIZ];
	int s, nread;

	(void) ioctl(fd, FIONREAD, &nread);

	printDBG(LVL_DEBUG, "::Old Client::\nCharacters Read: %d\n", nread);

	/* nothing read because client disappeared, close client connection */
	if (nread == 0) {
		if (pref.loud_disconnect)
			show_osd(cs_get_prompt(fd), "*** DISCONNECTED ***");

		srv_disconnect_client(fd);

		return;
	}

	/* warning level exceeded, close client connection */
	if (check_warnlevel(fd)) {
		printDBG(LVL_DEBUG, "Client has exceeded warning level...disconnecting.\n");
		show_osd(cs_get_prompt(fd), "*** Client exceeded Warning Level:  Disconnecting");

		srv_disconnect_client(fd);
		return;
	}

	/* read message from client */
	set_limits(&nread, 1, pref.wrapcol * xset.lines);
	(void) read(fd, msg, nread);

	msg[nread] = NIL;
	s = strlen(msg)-1;
	if (msg[s] == '\n')
		msg[s] = NIL;

	printDBG(LVL_DEBUG, "Message Length: %d\n", s+1);

	cs_set_tm(fd, time(NULL));

	/* adjust warning level if message was too long to fit in one view */
	if ( show_osd(cs_get_prompt(fd), msg) )
		cs_adj_warning(fd, 5);
}

/* Default signal handler */
void srv_set_quit(int sig)
{
	quit = sig;
}

/* Called when quit is set by default sig handler */
void srv_quit()
{
	printDBG(LVL_SYSLOG, "osdchatd "OC_VERSION" daemon terminated.  signal: %d.", quit);

	if (osd)
		xosd_destroy(osd);

	cs_del_close_all();
	cs_block_free();

	unlink(pid_fname);

	exit(EXIT_SUCCESS);
}

/* SIGHUP signal handler */
void srv_sighup(int sig)
{
	(void) sig;

	reload_config = 1;
	signal(SIGHUP, srv_sighup);
}

/* Called when reload_config is set during SIGHUP */
void srv_reload_config()
{
	printDBG(LVL_DEBUG, "Reloading config file\n");
	parse_config();

	if (osd) {
		printDBG(LVL_DEBUG, "Destroying current xosd object\n");
		xosd_destroy(osd);
		if ( !init_xosd() ) {
			quit = SIGHUP;
			srv_quit();
		}
	}

	reload_config = 0;
}

/* Kills currently running daemon */
void srv_kill(void)
{
	char s_pid[6];
	FILE *f;
	pid_t pid;

	if ( (f = safe_fopen(pid_fname, "r")) == NULL ) {
		if (errno == ENOENT)
			printDBG(LVL_FATERR, "osdchatd not running or pid file removed.\n");
		else
			printDBG(LVL_FATERR, "safe_fopen(): %m\n");
	}

	fgets(s_pid, 6, f);
	fclose(f);

	pid = (pid_t) atoi(s_pid);

	printf("Killing osdchatd daemon, pid=%d.\n", pid);
	if (kill(pid, SIGINT) == -1)
		perror("kill");

	unlink(pid_fname);
	exit(EXIT_FAILURE);
}

/* Make sure server does not get clobbered by too many quick messages
   Return non zero if warning level exceeded threshold.
*/
int check_warnlevel(int fd)
{
	int diff, warn, exceed = 0;
	time_t curr_t, last_t;

	curr_t = time(NULL);
	last_t = cs_get_tm(fd);
	printDBG(LVL_DEBUG, "curr: %ld\tlast: %ld\n", curr_t, last_t);

	if (last_t == -1) {
		printDBG(LVL_DEBUG, "cs_get_tm(): %m");
		return exceed;
	}

	diff = (int) difftime(curr_t, last_t);
	printDBG(LVL_DEBUG, "diff: %d\n", diff);
	warn = cs_adj_warning(fd, 0);

	if (diff == 0)
		warn = cs_adj_warning(fd, 1);
	else if (diff >= 3 && warn > 0)
		warn = cs_adj_warning(fd, -1);

	printDBG(LVL_DEBUG, "warn: %d\n", warn);

	if (warn >= pref.max_warnlevel) {
		exceed = 1;
		if (pref.blocklist)
			cs_block(fd);
	}

	return (exceed);
}

void log_msgs(char *line)
{
	char loc_path[BUFSIZE];
	FILE *f;

	if (!pref.logmsgs || user_info->pw_dir == NULL)
		return;

	snprintf(loc_path, sizeof(loc_path), "%s/%s", user_info->pw_dir, LOG_FILE);

	/* Even though the log file is in the user's home dir, it must still
	be opened in a safe manner.  Ex: Quit if blocking will occur because
	the log file already exists as a FIFO (and it's not being read).  No
	symlinks or FIFOs allowed */
	if ( (f = safe_fopen(loc_path, "a")) == NULL ) {
		printDBG(LVL_SYSLOG, "Cannot create log file %s: %m\n", loc_path);
		return;
	}

	fprintf(f, "%s\n", line);
	fclose(f);
}

/* Try to open a file stream in a more secure fashion.  This
should account for cases where the filesystem is changing
underneath you.
*/
FILE *safe_fopen(const char *fname, const char *mode)
{
        struct stat lstat_info;
        struct stat fstat_info;
        FILE *f = NULL;
        int fd = 0, temp_fd = 0, write = 0;

	write = (strchr(mode, 'w') || strchr(mode, 'a'));

	/* file doesn't exists */
        if ( lstat(fname, &lstat_info) == -1 ) {
		/* error other than file does not exist */
                if (errno != ENOENT)
                        printDBG(LVL_FATAL, "safe_fopen lstat(%s): %m\n", fname);

		/* if file open for writing, try to create it */
		if ( write && (fd = open(fname, O_CREAT | O_EXCL | O_RDWR, 0600)) == -1 )
                        printDBG(LVL_FATAL, "safe_fopen, open(): %m\n");

	/* file already exists */
        } else {
		if ( (temp_fd = open(fname, O_RDWR)) == -1 )
			printDBG(LVL_FATAL, "safe_fopen, open(%s): %m", fname);

		/* check if file is a link */
		if ( fstat(temp_fd, &fstat_info) == -1 || \
			lstat_info.st_mode != fstat_info.st_mode || \
			lstat_info.st_ino != fstat_info.st_ino || \
			lstat_info.st_dev != fstat_info.st_dev || \
			fstat_info.st_nlink > 1 )
		{
			printDBG(LVL_FATAL, "safe_fopen, fstat(%s):  Links are not allowed\n", fname);
		}

		/* check if the file is regular */
		if ( !S_ISREG(lstat_info.st_mode) )
			printDBG(LVL_FATAL, "safe_fopen, fstat(%s):  Not a regular file\n", fname);


		if ( (fd = open(fname, O_CREAT | O_RDWR)) == -1 )
			printDBG(LVL_FATAL, "safe_fopen, open(%s): %m\n", fname);
        }

        if ( fd != 0 && (f = fdopen(fd, mode)) == NULL) {
                printDBG(LVL_SYSLOG, "safe_fopen, fdopen(%s): %m\n", fname);
                unlink(fname);
                exit(EXIT_FAILURE);    /* internal error, should never occur */
        }

	if (temp_fd != 0)
		close(temp_fd);

        return f;
}
